<?php
/**
 * FuzeWorks Authentication Plugin.
 *
 * The FuzeWorks PHP FrameWork
 *
 * Copyright (C) 2013 - 2023 i15
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * @author    i15
 * @copyright Copyright (C) 2013 - 2023, i15. (https://i15.nl)
 * @license   https://opensource.org/licenses/MIT MIT License
 *
 * @since Version 1.3.0
 *
 * @version Version 1.3.0
 */

namespace FuzeWorks\Authentication\Drivers;

use FuzeWorks\Authentication\Model\Session;
use FuzeWorks\Authentication\Model\Users;
use FuzeWorks\Authentication\SessionsModelDriverInterface;
use FuzeWorks\DatabaseEngine\iDatabaseEngine;
use FuzeWorks\DatabaseEngine\MongoEngine;
use MongoDB\Collection;

class MongoSessionsModelDriver implements SessionsModelDriverInterface
{

    /** @var MongoEngine  */
    protected iDatabaseEngine $engine;
    protected Collection $collection;
    protected Users $usersModel;

    public function __construct(iDatabaseEngine $engine, Users $usersModel)
    {
        $this->engine = $engine;
        $this->collection = $this->engine->selectDatabase("auth")->selectCollection("sessions");
        $this->usersModel = $usersModel;
    }

    /**
     * @param Session $session
     * @return bool
     */
    public function createSession(Session $session): bool
    {
        $insert = [
            "sessionKey" => $session->sessionKey,
            "user" => $session->user->id,
            "expiryDate" => $session->expiryDate,
            "expiryThreshold" => $session->expiryThreshold,
            "persistent" => $session->persistent,
            "active" => $session->active,
            "context" => $session->context
        ];

        $res = $this->collection->insertOne($insert);
        if ($res->getInsertedCount() !== 1)
            return false;

        return true;
    }

    public function readSessions(array $filter): array
    {
        // Fetch from database
        $results = $this->collection->find($filter);
        $out = [];
        foreach ($results as $result)
        {
            // Load variables
            $sessionKey = $result->sessionKey;
            $user = $this->usersModel->getUserByUid($result->user);
            $expiryDate = $result->expiryDate;
            $expiryThreshold = $result->expiryThreshold;
            $persistent = $result->persistent;
            $active = $result->active;
            $context = (array) $result->context;

            $out[] = new Session($user, $sessionKey, $expiryDate, $expiryThreshold, $context, $persistent, $active);
        }

        // And output cumulative results
        return $out;
    }

    public function updateSessions(array $sessions): bool
    {
        $modified = 0;
        foreach ($sessions as $session)
        {
            $update = [
                "user" => $session->user->id,
                "expiryDate" => $session->expiryDate,
                "expiryThreshold" => $session->expiryThreshold,
                "persistent" => $session->persistent,
                "active" => $session->active,
                "context" => $session->context
            ];

            $res = $this->collection->updateOne(
                ["sessionKey" => $session->sessionKey],
                ['$set' => $update]
            );
            $modified += $res->getModifiedCount();
        }

        if ($modified === count($sessions))
            return true;

        return false;
    }

    public function deleteSessions(array $sessions): bool
    {
        $deleted = 0;
        foreach ($sessions as $session)
        {
            $res = $this->collection->deleteOne(["sessionKey" => $session->sessionKey]);
            $deleted += $res->getDeletedCount();
        }

        if ($deleted === count($sessions))
            return true;

        return false;
    }
}